# Copyright 2019-2020 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections.abc
import types
import typing
from collections.abc import Callable, Iterable
from enum import Enum  # noqa: F401 (doctest)
from http import HTTPStatus
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Literal,
    Optional,
    TypeVar,
    Union,
    get_args,
    get_origin,
)
from uuid import UUID

import strawberry
from annotated_types import Len, MaxLen, MinLen
from more_itertools import first, last
from pydantic.fields import FieldInfo

from pydantic_forms.types import InputForm, State, strEnum

__all__ = [
    "SAFE_USED_BY_TRANSITIONS_FOR_STATUS",
    "BroadcastFunc",
    "ErrorDict",
    "ErrorState",
    "StateStepFunc",
    "StepFunc",
    "SubscriptionLifecycle",
    "filter_nonetype",
    "get_origin_and_args",
    "get_possible_product_block_types",
    "is_list_type",
    "is_of_type",
    "is_optional_type",
    "is_union_type",
    "list_factory",
]

if TYPE_CHECKING:
    from orchestrator.domain.base import ProductBlockModel


def is_union(tp: type[Any] | None) -> bool:
    return tp is Union or tp is types.UnionType  # type: ignore[comparison-overlap]


# ErrorState is either a string containing an error message, a caught Exception, or a tuple containing a message and
# an HTTP status code
ErrorState = Union[str, Exception, tuple[str, Union[int, HTTPStatus]]]
# An ErrorDict should have the following keys:
# error: str  # A message describing the error
# class: str[Optional]  # The exception class name (type)
# status_code: Optional[int]  # HTTP Status code (optional)
# traceback: Optional[str]  # A python traceback as a string formatted by nwastdlib.ex.show_ex
ErrorDict = dict[str, Union[str, int, list[dict[str, Any]], InputForm, None]]
StateStepFunc = Callable[[State], State]
StepFunc = Callable[..., Optional[State]]
BroadcastFunc = Callable[[UUID], None]

SI = TypeVar("SI")


@strawberry.enum
class SubscriptionLifecycle(strEnum):
    INITIAL = "initial"
    ACTIVE = "active"
    MIGRATING = "migrating"
    DISABLED = "disabled"
    TERMINATED = "terminated"
    PROVISIONING = "provisioning"


# TODO #1321: old code that protected against unsafe changes in subs
# The key is the Parent subscription life cycle status. The keys are lists of safe transitions for child subscriptions.
SAFE_USED_BY_TRANSITIONS_FOR_STATUS = {
    SubscriptionLifecycle.INITIAL: [
        SubscriptionLifecycle.INITIAL,
    ],
    SubscriptionLifecycle.ACTIVE: [
        SubscriptionLifecycle.INITIAL,
        SubscriptionLifecycle.PROVISIONING,
        SubscriptionLifecycle.MIGRATING,
        SubscriptionLifecycle.ACTIVE,
        SubscriptionLifecycle.TERMINATED,
        SubscriptionLifecycle.DISABLED,
    ],
    SubscriptionLifecycle.MIGRATING: [
        SubscriptionLifecycle.INITIAL,
        SubscriptionLifecycle.MIGRATING,
        SubscriptionLifecycle.TERMINATED,
    ],
    SubscriptionLifecycle.PROVISIONING: [
        SubscriptionLifecycle.INITIAL,
        SubscriptionLifecycle.PROVISIONING,
        SubscriptionLifecycle.ACTIVE,
        SubscriptionLifecycle.TERMINATED,
    ],
    SubscriptionLifecycle.TERMINATED: [SubscriptionLifecycle.INITIAL, SubscriptionLifecycle.TERMINATED],
    SubscriptionLifecycle.DISABLED: [
        SubscriptionLifecycle.INITIAL,
        SubscriptionLifecycle.DISABLED,
        SubscriptionLifecycle.TERMINATED,
    ],
}


def is_of_type(t: Any, test_type: Any) -> bool:
    """Check if annotation type is valid for type.

    >>> is_of_type(list, list)
    True
    >>> is_of_type(list[int], list[int])
    True
    >>> is_of_type(strEnum, str)
    True
    >>> is_of_type(strEnum, Enum)
    True
    >>> is_of_type(int, str)
    False
    >>> is_of_type(Any, Any)
    True
    >>> is_of_type(Any, int)
    True
    """
    if t is Any:
        return True

    if is_union_type(test_type):
        return any(get_origin(t) is get_origin(arg) for arg in get_args(test_type))

    if (
        get_origin(t)
        and get_origin(test_type)
        and get_origin(t) is get_origin(test_type)
        and get_args(t) == get_args(test_type)
    ):
        return True

    if test_type is t:
        # Test type is a typing type instance and matches
        return True

    # Workaround for the fact that you can't call issubclass on typing types
    try:
        return issubclass(t, test_type)
    except TypeError:
        return False


# TODO #1330: Fix comparison of is_of_type(Union[str, int], Union[int, str]), it returns False but it should be True
def is_list_type(t: Any, test_type: type | None = None) -> bool:
    """Check if `t` is list type.

    And optionally check if the list items are of `test_type`

    >>> is_list_type(list[int])
    True
    >>> is_list_type(list[Any], int)
    True
    >>> is_list_type(list, int)
    False
    >>> is_list_type(Optional[list[int]])
    True
    >>> is_list_type(Optional[list[int]], int)
    True
    >>> is_list_type(Annotated[Optional[list[int]], "foo"], int)
    True
    >>> is_list_type(Optional[list[int]], str)
    False
    >>> is_list_type(Annotated[Optional[list[int]], "foo"], str)
    False
    >>> is_list_type(Optional[int])
    False
    >>> is_list_type(list[tuple[int, int]])
    True
    >>> is_list_type(list[tuple[int, int]], int)
    False
    >>> is_list_type(list[tuple[int, int]], tuple[int, int])
    True
    >>> is_list_type(list[strEnum], Enum)
    True
    >>> is_list_type(int)
    False
    >>> is_list_type(Literal[1,2,3])
    False
    >>> is_list_type(list[Union[str, int]])
    True
    >>> is_list_type(list[Union[str, int]], Union[str, int])
    False
    >>> is_list_type(list[Union[str, int]], str)
    True
    >>> is_list_type(list[Union[str, int]], int)
    False
    >>> is_list_type(list[Union[str, int]], Union[int, int])
    False
    >>> from pydantic import conlist
    >>> is_list_type(conlist(int))
    True
    >>> is_list_type(Annotated[list, "foo"])
    True
    >>> is_list_type(Annotated[list, "foo"], int)
    False
    >>> is_list_type(Annotated[list[str], "foo"], int)
    False
    >>> is_list_type(Annotated[list[str], "foo"], str)
    True
    >>> from typing import Sequence
    >>> is_list_type(Annotated[Sequence[str], "foo"], str)
    True
    >>> is_list_type({"foo": "bar"})
    False
    >>> is_list_type((1, 2, 3))
    False
    >>> is_list_type({1, 2, 3})
    False
    """
    t_origin, t_args = get_origin_and_args(t)
    if t_origin is None:
        return test_type is None and has_list_in_mro(t)

    if is_optional_type(t) or is_union_type(t):
        for arg in t_args:
            if is_list_type(arg, test_type):
                return True
    elif t_origin == Literal:
        return False  # Literal cannot contain lists see pep 586
    elif issubclass(t_origin, list) or t_origin in (collections.abc.Sequence, typing.Sequence):
        if test_type and t_args:
            first_arg = first(t_args)
            # To support a list/sequence with union of multiple product blocks.
            if is_union_type(first_arg) and get_args(first_arg) and not is_union_type(test_type):
                first_arg = get_args(first_arg)[0]
            return is_of_type(first_arg, test_type)
        return True

    return False


def get_origin_and_args(t: Any) -> tuple[Any, tuple[Any, ...]]:
    """Return the origin and args of the given type.

    When wrapped in Annotated[] this is removed.
    """
    origin, args = get_origin(t), get_args(t)
    if origin is not Annotated:
        return origin, args

    t_unwrapped = first(args)
    return get_origin(t_unwrapped), get_args(t_unwrapped)


def filter_nonetype(types_: Iterable[Any]) -> Iterable[Any]:
    def not_nonetype(type_: Any) -> bool:
        return type_ is not None.__class__

    return filter(not_nonetype, types_)


def is_optional_type(t: Any, test_type: type | None = None) -> bool:
    """Check if `t` is optional type (Union[None, ...]).

    And optionally check if T is of `test_type`

    >>> is_optional_type(Optional[int])
    True
    >>> is_optional_type(Annotated[Optional[int], "foo"])
    True
    >>> is_optional_type(Annotated[int, "foo"])
    False
    >>> is_optional_type(Union[None, int])
    True
    >>> is_optional_type(Union[int, str, None])
    True
    >>> is_optional_type(Union[int, str])
    False
    >>> is_optional_type(Optional[int], int)
    True
    >>> is_optional_type(Optional[int], str)
    False
    >>> is_optional_type(Annotated[Optional[int], "foo"], int)
    True
    >>> is_optional_type(Annotated[Optional[int], "foo"], str)
    False
    >>> is_optional_type(Optional[State], int)
    False
    >>> is_optional_type(Optional[State], State)
    True
    """
    origin, args = get_origin_and_args(t)

    if is_union(origin) and None.__class__ in args:
        field_type = first(filter_nonetype(args))
        return test_type is None or is_of_type(field_type, test_type)
    return False


def is_union_type(t: Any, test_type: type | None = None) -> bool:
    """Check if `t` is union type (Union[Type, AnotherType]).

    Optionally check if T is of `test_type` We cannot check for literal Nones.

    >>> is_union_type(Union[int, str])
    True
    >>> is_union_type(Annotated[Union[int, str], "foo"])
    True
    >>> is_union_type(Union[int, str], str)
    True
    >>> is_union_type(Union[int, str], bool)
    False
    >>> is_union_type(Union[int, str], Union[int, str])
    True
    >>> is_union_type(Union[int, None])
    True
    >>> is_union_type(Annotated[Union[int, None], "foo"])
    True
    >>> is_union_type(int)
    False
    """
    origin, args = get_origin_and_args(t)
    if not is_union(origin):
        return False
    if not test_type:
        return True

    if is_of_type(t, test_type):
        return True

    for arg in args:
        result = is_of_type(arg, test_type)
        if result:
            return result
    return False


def get_possible_product_block_types(
    list_field_type: type["ProductBlockModel"] | type["ProductBlockModel"],
) -> dict[str, type["ProductBlockModel"]]:
    _origin, list_item_field_type_args = get_origin_and_args(list_field_type)
    if not is_union_type(list_field_type):
        return {list_field_type.name: list_field_type} if list_field_type.name else {}

    possible_product_block_types = {}
    for list_item_field_type in filter_nonetype(list_item_field_type_args):
        if list_item_field_type.name not in possible_product_block_types:
            possible_product_block_types[list_item_field_type.name] = list_item_field_type
    return possible_product_block_types


def has_list_in_mro(type_: Any) -> bool:
    """Check if the type is (derived from) list.

    >>> has_list_in_mro(list)
    True
    >>> has_list_in_mro(type("ListSubclass", (list,), {}))
    True
    >>> has_list_in_mro(tuple)
    False
    >>> has_list_in_mro(dict)
    False
    """
    try:
        return list in type_.mro()
    except AttributeError:
        return False


def yield_min_length(type_args: Iterable) -> Iterable[int]:
    """Given an iterable of type args, yield min_length values found in typing metadata."""
    for type_arg in type_args:
        if isinstance(type_arg, (Len, MinLen)):
            yield type_arg.min_length
        if isinstance(type_arg, FieldInfo):
            yield from yield_min_length(type_arg.metadata)


def yield_max_length(type_args: Iterable) -> Iterable[int]:
    """Given an iterable of type args, yield max_length values found in typing metadata."""
    for type_arg in type_args:
        if isinstance(type_arg, (Len, MaxLen)) and type_arg.max_length is not None:
            yield type_arg.max_length
        if isinstance(type_arg, FieldInfo):
            yield from yield_max_length(type_arg.metadata)


def get_iterable_max_length(iterable_type: Any, reduce_func: Callable = last) -> int:
    """Return the max_length for the given annotated iterable type.

    Args:
        iterable_type: the iterable type (i.e. list, Sequence)
        reduce_func: callable to reduce multiple max_length values to one value.
            Defaults to `more_itertools.last`
    """
    return reduce_func(yield_max_length(get_args(iterable_type)))


def _get_default_type(type_: Any) -> Any:
    """Given a type, return the default type."""
    type_ = first(get_args(type_))
    if is_union_type(type_):
        # We return the last not-none type from the union (same as before in _init_instances())
        # Note that:
        # > If X is a union [..] the order [..] may be different from the order of the original arguments
        # https://docs.python.org/3.11/library/typing.html#typing.get_args
        type_ = last(filter_nonetype(get_args(type_)))
    return type_


def list_factory(type_: Any, *init_args: Any, **init_kwargs: Any) -> list:
    """Given a list type, create a list with optionally values.

    Args:
        type_: type of the list elements, is searched for min_length annotation
        init_args: positional args passed to created elements
        init_kwargs: keyword args passed to created elements

    >>> from annotated_types import Len
    >>> from pydantic import conlist, Field
    >>> from typing import get_args, Annotated, Sequence
    >>> list_factory(list)
    []
    >>> list_factory(list[int])
    []
    >>> list_factory(conlist(int))
    []
    >>> list_factory(Annotated[list[int], Field(min_length=2)], 1)
    [1, 1]
    >>> list_factory(Annotated[list, Field(min_length=2)], 1)
    []
    >>> list_factory(Annotated[list[int], Len(min_length=2)], 1)
    [1, 1]
    >>> list_factory(conlist(int, min_length=2), 1)
    [1, 1]
    >>> list_factory(conlist(Union[int, str], min_length=2), 1) in ([1, 1], ['1', '1'])
    True
    >>> list_factory(conlist(dict, min_length=1), foo="bar")
    [{'foo': 'bar'}]
    >>> list_factory(Annotated[Sequence[int], Len(min_length=1)], 3)
    [3]
    """
    from orchestrator.domain.base import ProductBlockModel

    if get_origin(type_) in (None, list):
        # List without annotations
        return []

    base_type, *type_args = get_args(type_)
    if base_type is list:
        # List with annotations but without a type
        return []

    if not is_list_type(base_type):
        raise TypeError(f"Tried to initialize a list for a non-list pb field: {type_=}")

    # Look for a type_arg specifying a minimum length
    length = first(yield_min_length(type_args), None)
    # TODO when someone specifies min_length multiple times with different values, should we do something?
    if not length:
        return []

    def make_default_value(t: Any) -> Any:
        if issubclass(t, ProductBlockModel):
            return t.new(*init_args, **init_kwargs)

        # Support for non-ProductBlockModel types is mainly for doctests, we could remove it
        return t(*init_args, **init_kwargs)

    default_type = _get_default_type(base_type)
    return [make_default_value(default_type) for _ in range(length)]



"""Validate Product Type.

Revision ID: 4fjdn13f83ga
Revises: 2c7e8a43d4f9
Create Date: 2025-10-13 16:21:43.956814

"""

from uuid import uuid4

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "4fjdn13f83ga"
down_revision = "4c5859620539"
branch_labels = None
depends_on = None


workflow = {
    "name": "task_validate_product_type",
    "target": "SYSTEM",
    "description": "Validate all subscriptions of Product Type",
    "workflow_id": uuid4(),
}


def upgrade() -> None:
    conn = op.get_bind()
    conn.execute(
        sa.text(
            "INSERT INTO workflows VALUES (:workflow_id, :name, :target, :description, now()) ON CONFLICT DO NOTHING"
        ),
        workflow,
    )


def downgrade() -> None:
    conn = op.get_bind()
    conn.execute(sa.text("DELETE FROM workflows WHERE name = :name"), {"name": workflow["name"]})



import structlog
from sqlalchemy import BinaryExpression

from orchestrator.db import ProductBlockTable, ResourceTypeTable
from orchestrator.db.filters import create_memoized_field_list, generic_filter_from_clauses
from orchestrator.db.filters.search_filters import default_inferred_column_clauses, node_to_str_val
from orchestrator.utils.search_query import Node

logger = structlog.get_logger(__name__)


def product_blocks_clause(node: Node) -> BinaryExpression:
    return ResourceTypeTable.product_blocks.any(ProductBlockTable.name.ilike(node_to_str_val(node)))


RESOURCE_TYPE_TABLE_COLUMN_CLAUSES = default_inferred_column_clauses(ResourceTypeTable) | {
    "product_block": product_blocks_clause,
}

resource_type_filter_fields = create_memoized_field_list(RESOURCE_TYPE_TABLE_COLUMN_CLAUSES)
filter_resource_types = generic_filter_from_clauses(RESOURCE_TYPE_TABLE_COLUMN_CLAUSES)



"""Search index model for llm integration.

Revision ID: 52b37b5b2714
Revises: 850dccac3b02
Create Date: 2025-08-12 22:34:26.694750

"""

import sqlalchemy as sa
from alembic import op
from pgvector.sqlalchemy import Vector
from sqlalchemy.dialects import postgresql
from sqlalchemy_utils import LtreeType

from orchestrator.search.core.types import FieldType

# revision identifiers, used by Alembic.
revision = "52b37b5b2714"
down_revision = "850dccac3b02"
branch_labels = None
depends_on = None

TABLE = "ai_search_index"
IDX_EMBED_HNSW = "ix_flat_embed_hnsw"
IDX_PATH_GIST = "ix_flat_path_gist"
IDX_PATH_BTREE = "ix_flat_path_btree"
IDX_VALUE_TRGM = "ix_flat_value_trgm"
IDX_CONTENT_HASH = "idx_ai_search_index_content_hash"

TARGET_DIM = 1536


def upgrade() -> None:
    # Create PostgreSQL extensions
    op.execute("CREATE EXTENSION IF NOT EXISTS ltree;")
    op.execute("CREATE EXTENSION IF NOT EXISTS unaccent;")
    op.execute("CREATE EXTENSION IF NOT EXISTS pg_trgm;")
    op.execute("CREATE EXTENSION IF NOT EXISTS vector;")

    # Create the ai_search_index table
    op.create_table(
        TABLE,
        sa.Column("entity_type", sa.Text, nullable=False),
        sa.Column("entity_id", postgresql.UUID, nullable=False),
        sa.Column("path", LtreeType, nullable=False),
        sa.Column("value", sa.Text, nullable=False),
        sa.Column("embedding", Vector(TARGET_DIM), nullable=True),
        sa.Column("content_hash", sa.String(64), nullable=False),
        sa.PrimaryKeyConstraint("entity_id", "path", name="pk_ai_search_index"),
    )

    field_type_enum = sa.Enum(*[ft.value for ft in FieldType], name="field_type")
    field_type_enum.create(op.get_bind(), checkfirst=True)
    op.add_column(
        TABLE,
        sa.Column("value_type", field_type_enum, nullable=False, server_default=FieldType.STRING.value),
    )
    op.alter_column(TABLE, "value_type", server_default=None)

    op.create_index(op.f("ix_ai_search_index_entity_id"), TABLE, ["entity_id"], unique=False)
    op.create_index(IDX_CONTENT_HASH, TABLE, ["content_hash"])

    op.create_index(
        IDX_PATH_GIST,
        TABLE,
        ["path"],
        postgresql_using="GIST",
        postgresql_ops={"path": "gist_ltree_ops"},
    )
    op.create_index(IDX_PATH_BTREE, TABLE, ["path"])
    op.create_index(IDX_VALUE_TRGM, TABLE, ["value"], postgresql_using="GIN", postgresql_ops={"value": "gin_trgm_ops"})

    op.create_index(
        IDX_EMBED_HNSW,
        TABLE,
        ["embedding"],
        postgresql_using="HNSW",
        postgresql_with={"m": 16, "ef_construction": 64},
        postgresql_ops={"embedding": "vector_l2_ops"},
    )


def downgrade() -> None:
    # Drop all indexes
    op.drop_index(IDX_EMBED_HNSW, table_name=TABLE, if_exists=True)
    op.drop_index(IDX_VALUE_TRGM, table_name=TABLE, if_exists=True)
    op.drop_index(IDX_PATH_BTREE, table_name=TABLE, if_exists=True)
    op.drop_index(IDX_PATH_GIST, table_name=TABLE, if_exists=True)
    op.drop_index(IDX_CONTENT_HASH, table_name=TABLE, if_exists=True)
    op.drop_index(op.f("ix_ai_search_index_entity_id"), table_name=TABLE, if_exists=True)

    # Drop table and enum
    op.drop_table(TABLE, if_exists=True)
    field_type_enum = sa.Enum(name="field_type")
    field_type_enum.drop(op.get_bind(), checkfirst=True)



# Copyright 2019-2020 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Iterable, Sequence
from typing import Any
from uuid import UUID

import sqlalchemy as sa
import structlog

from orchestrator.settings import app_settings
from pydantic_forms.types import UUIDstr

logger = structlog.get_logger(__name__)


def has_table_column(table_name: str, column_name: str, conn: sa.engine.Connection) -> bool:
    """Checks if the specified column exists in a given table.

    inspector.get_columns raises an exception if the table does not exist, so we catch that exception and return False.
    This is useful for migrations where you want to ensure that a column exists before performing operations on it.

    :param table_name: Name of the database table
    :param column_name: Name of the column to check
    :param conn: SQLAlchemy database Connection
    :return: True if the column exists, False otherwise
    """
    result = conn.execute(
        sa.text(
            """
            SELECT column_name
            FROM information_schema.columns
            WHERE table_name = :table_name and column_name = :column_name
            """
        ),
        {
            "table_name": table_name,
            "column_name": column_name,
        },
    )
    return result.first() is not None


def get_resource_type_id_by_name(conn: sa.engine.Connection, name: str) -> UUID:
    result = conn.execute(
        sa.text("SELECT resource_type_id FROM resource_types WHERE resource_types.resource_type=:name"), {"name": name}
    )
    return [x for (x,) in result.fetchall()][0]


def get_fixed_input_id_by_name(conn: sa.engine.Connection, name: str) -> UUID:
    result = conn.execute(
        sa.text("SELECT fixed_input_id FROM fixed_inputs WHERE fixed_inputs.name=:name"), {"name": name}
    )
    return [x for (x,) in result.fetchall()][0]


def get_product_block_id_by_name(conn: sa.engine.Connection, name: str) -> UUID:
    result = conn.execute(
        sa.text("SELECT product_block_id FROM product_blocks WHERE product_blocks.name=:name"), {"name": name}
    )
    return [x for (x,) in result.fetchall()][0]


def get_product_id_by_name(conn: sa.engine.Connection, name: str) -> UUID:
    result = conn.execute(sa.text("SELECT product_id FROM products WHERE products.name=:name"), {"name": name})
    return [x for (x,) in result.fetchall()][0]


def get_product_name_by_id(conn: sa.engine.Connection, id: UUID | UUIDstr) -> str:
    result = conn.execute(sa.text("SELECT name FROM products WHERE product_id=:id"), {"id": id})
    return [x for (x,) in result.fetchall()][0]


def get_product_by_id(conn: sa.engine.Connection, id: UUID | UUIDstr) -> dict:
    result = conn.execute(sa.text("SELECT * FROM products WHERE product_id=:id"), {"id": id})

    return dict(result.mappings().one())
    # return result.fetchall()[0]


def get_product_ids_by_product_type(conn: sa.engine.Connection, product_type: str) -> list[UUID]:
    result = conn.execute(
        sa.text("SELECT product_id FROM products WHERE product_type=:product_type"),
        {"product_type": product_type},
    )
    return [x for (x,) in result.fetchall()]


def get_fixed_inputs_by_product_id(conn: sa.engine.Connection, id: UUID | UUIDstr) -> Sequence:
    result = conn.execute(sa.text("SELECT * FROM fixed_inputs WHERE product_id=:id"), {"id": id})
    return result.fetchall()


def insert_resource_type(conn: sa.engine.Connection, resource_type: str, description: str) -> None:
    """Create a new resource types."""
    conn.execute(
        sa.text(
            """INSERT INTO resource_types (resource_type, description)
               VALUES (:resource_type, :description)
               ON CONFLICT DO NOTHING;"""
        ),
        {
            "resource_type": resource_type,
            "description": description,
        },
    )


def get_all_active_products_and_ids(conn: sa.engine.Connection) -> list[dict[str, UUID | UUIDstr | str]]:
    """Return a list, with dicts containing keys `product_id` and `name` of active products."""
    result = conn.execute(sa.text("SELECT product_id, name  FROM products WHERE status='active'"))
    return [{"product_id": row[0], "name": row[1]} for row in result.fetchall()]


def ensure_default_workflows(conn: sa.engine.Connection) -> None:
    """Ensure products_workflows table contains a link between all 'active' workflows and the set of workflows identified in the DEFAULT_PRODUCT_WORKFLOWS app_setting.

    Note that the 0th element of the uuids are taken when generating product_workflow_table_rows because sqlalchemy returns a row tuple even if selecting for a single column.
    """
    products = sa.Table("products", sa.MetaData(), autoload_with=conn)
    workflows = sa.Table("workflows", sa.MetaData(), autoload_with=conn)
    product_workflows_table = sa.Table("products_workflows", sa.MetaData(), autoload_with=conn)

    all_product_uuids = conn.execute(sa.select(products.c.product_id)).fetchall()
    default_workflow_ids = conn.execute(
        sa.select(workflows.c.workflow_id).where(workflows.c.name.in_(app_settings.DEFAULT_PRODUCT_WORKFLOWS))
    ).fetchall()
    product_workflow_table_rows = [
        (product_uuid[0], workflow_uuid[0])
        for product_uuid in all_product_uuids
        for workflow_uuid in default_workflow_ids
    ]
    conn.execute(
        sa.dialects.postgresql.insert(product_workflows_table)
        .values(product_workflow_table_rows)
        .on_conflict_do_nothing(index_elements=("product_id", "workflow_id"))
    )


def create_workflow(conn: sa.engine.Connection, workflow: dict) -> None:
    """Create a new workflow for a product.

    Args:
        conn: DB connection as available in migration main file.
        workflow: Dict with data for a new workflow.
            name: Name of the workflow.
            target: Target of the workflow ("CREATE", "MODIFY", "RECONCILE", "TERMINATE", "SYSTEM")
            description: Description of the workflow.
            product_type: Product type to add the workflow to.

    Usage:
        >>> workflow = {
            "name": "workflow_name",
            "target": "SYSTEM",
            "is_task": False,
            "description": "workflow description",
            "product_type": "product_type",
        }
        >>> create_workflow(conn, workflow)
    """
    if not workflow.get("is_task", False):
        workflow["is_task"] = False

    if has_table_column(table_name="workflows", column_name="is_task", conn=conn):
        query = """
                WITH new_workflow AS (
                    INSERT INTO workflows (name, target, is_task, description)
                        VALUES (:name, :target, :is_task, :description)
                        ON CONFLICT DO NOTHING
                        RETURNING workflow_id)
                INSERT
                INTO products_workflows (product_id, workflow_id)
                SELECT p.product_id,
                       nw.workflow_id
                FROM products AS p
                         CROSS JOIN new_workflow AS nw
                WHERE p.product_type = :product_type
                ON CONFLICT DO NOTHING
                """
    else:
        # Remove is_task from workflow dict and insert SQL
        workflow = {k: v for k, v in workflow.items() if k != "is_task"}
        query = """
                WITH new_workflow AS (
                    INSERT INTO workflows (name, target, description)
                        VALUES (:name, :target, :description)
                        ON CONFLICT DO NOTHING
                        RETURNING workflow_id)
                INSERT
                INTO products_workflows (product_id, workflow_id)
                SELECT p.product_id, nw.workflow_id
                FROM products AS p
                         CROSS JOIN new_workflow AS nw
                WHERE p.product_type = :product_type
                ON CONFLICT DO NOTHING
                """

    conn.execute(sa.text(query), workflow)


def create_task(conn: sa.engine.Connection, task: dict) -> None:
    """Create a new task.

    Args:
        conn: DB connection as available in migration main file.
        task: Dict with data for a new workflow.
            name: Name of the task.
            description: Description of the workflow.

    Usage:
        >>> task = {
            "name": "task_name",
            "description": "task description",
        }
        >>> create_workflow(conn, task)
    """
    if has_table_column(table_name="workflows", column_name="is_task", conn=conn):
        query = """
                INSERT INTO workflows(name, target, is_task, description)
                VALUES (:name, 'SYSTEM', TRUE, :description)
                ON CONFLICT DO NOTHING
                RETURNING workflow_id
                """
    else:
        query = """
                INSERT INTO workflows(name, target, description)
                VALUES (:name, 'SYSTEM', :description)
                ON CONFLICT DO NOTHING
                RETURNING workflow_id
                """

    conn.execute(sa.text(query), task)


def create_workflows(conn: sa.engine.Connection, new: dict) -> None:
    """Create new workflows.

    Args:
        conn: DB connection as available in migration main file
        new: an dict of your workflow data

    Example:
        >>> new_workflows = {
                "workflow_name": {
                    "workflow_id": "f2702074-3203-454c-b298-6dfa7675423d",
                    "target": "CREATE",
                    "is_task": False,
                    "description": "Workflow description",
                    "tag": "ProductBlockName1",
                    "search_phrase": "Search Phrase%",
                }
            }
    """
    for name, workflow in new.items():
        workflow["name"] = name

        if not workflow.get("is_task", False):
            workflow["is_task"] = False

        if has_table_column(table_name="workflows", column_name="is_task", conn=conn):
            query = """
                    WITH new_workflow AS (
                        INSERT INTO workflows (workflow_id, name, target, is_task, description)
                            VALUES (:workflow_id, :name, :target, :is_task, :description)
                            RETURNING workflow_id)
                    INSERT
                    INTO products_workflows (product_id, workflow_id)
                    SELECT p.product_id,
                           nw.workflow_id
                    FROM products AS p
                             CROSS JOIN new_workflow AS nw
                    WHERE p.tag = :tag
                      AND p.name LIKE :search_phrase
                    """
        else:
            # Remove is_task from workflow dict and insert SQL
            workflow = {k: v for k, v in workflow.items() if k != "is_task"}
            query = """
                    WITH new_workflow AS (
                        INSERT INTO workflows (workflow_id, name, target, description)
                            VALUES (:workflow_id, :name, :target, :description)
                            RETURNING workflow_id)
                    INSERT
                    INTO products_workflows (product_id, workflow_id)
                    SELECT p.product_id,
                           nw.workflow_id
                    FROM products AS p
                             CROSS JOIN new_workflow AS nw
                    WHERE p.tag = :tag
                      AND p.name LIKE :search_phrase
                    """

        conn.execute(sa.text(query), workflow)


def create_fixed_inputs(conn: sa.engine.Connection, product_id: UUID | UUIDstr, new: dict) -> dict[str, str]:
    """Create fixed inputs for a given product.

    Args:
        conn: DB connection as available in migration main file
        product_id: UUID of the product to link to
        new: an dict of your workflow data

    Example:
        >>> product_id = "id"
        >>> new = {
                "fixed_input_1": ("value", "f6a4f529-ad17-4ad8-b8ba-45684e2354ba"),
                "fixed_input_2": ("value", "5a67321d-45d5-4921-aa93-b8708b5d74c6")
            }
        >>> create_resource_types(conn, product_id, new)

    without extra ID's you don't need the tuple:

        >>> product_id = "id"
        >>> new = {
            "fixed_input_1": "value",
            "fixed_input_2": "value",
        }
        >>> create_fixed_inputs(conn, product_id, new)

    """
    insert_fixed_input_with_id = sa.text(
        """INSERT INTO fixed_inputs (fixed_input_id, name, value, created_at, product_id)
           VALUES (:fixed_input_id, :key, :value, now(), :product_id)
           ON CONFLICT DO NOTHING;"""
    )
    insert_fixed_input_without_id = sa.text(
        """INSERT INTO fixed_inputs (name, value, created_at, product_id)
           VALUES (:key, :value, now(), :product_id)
           ON CONFLICT DO NOTHING;"""
    )

    uuids = {}
    for key, values in new.items():
        if isinstance(values, tuple):
            value, fixed_input_id = values
            uuids[key] = fixed_input_id
            conn.execute(
                insert_fixed_input_with_id,
                {"fixed_input_id": fixed_input_id, "key": key, "value": value, "product_id": product_id},
            )
        else:
            conn.execute(insert_fixed_input_without_id, {"key": key, "value": values, "product_id": product_id})
            uuids[key] = get_fixed_input_id_by_name(conn, key)

    return uuids


def create_products(conn: sa.engine.Connection, new: dict) -> dict[str, UUIDstr]:
    """Create new products with their fixed inputs.

    Args:
        conn: DB connection as available in migration main file
        new: an dict of your workflow data

    Example:
        >>> new = {
                "Example Product": {
                    "product_id": "c9dc2374-514c-11eb-b685-acde48001122",
                    "product_type": "ProductType1",
                    "description": "Product description",
                    "tag": "ProductType",
                    "status": "active",
                    "fixed_inputs": {
                        "fixed_input_1": "value",
                        "fixed_input_2": "value2"
                    }
                },
                "Example Product 2": {
                    "product_type": "ProductType1",
                    "description": "Product description",
                    "tag": "ProductType",
                    "status": "active",
                    "product_block_ids": [
                        "37afe017-5a04-4d87-96b0-b8f88a328d7a"
                    ]
                }
            }
    """
    uuids = {}
    for name, product in new.items():
        product["name"] = name
        current_uuid = product["product_id"]
        uuids[name] = current_uuid
        conn.execute(
            sa.text(
                """
                INSERT INTO products (product_id, name, description, product_type, tag, status, created_at)
                VALUES (:product_id, :name, :description, :product_type, :tag, :status, now())
                ON CONFLICT DO NOTHING;
                """
            ),
            product,
        )
        for product_block_uuid in product.get("product_block_ids", []):
            conn.execute(
                sa.text("INSERT INTO product_product_blocks VALUES (:product_id, :product_block_id)"),
                {
                    "product_id": current_uuid,
                    "product_block_id": product_block_uuid,
                },
            )
        if "fixed_inputs" in product:
            create_fixed_inputs(conn, current_uuid, product["fixed_inputs"])
    return uuids


def create_product_blocks(conn: sa.engine.Connection, new: dict) -> dict[str, UUIDstr]:
    """Create new product blocks.

    Args:
        conn: DB connection as available in migration main file
        new: an dict of your workflow data
        products: list of product block ids to link these product blocks to

    Example:
        >>> new = {
                "Example Product Block": {
                    "product_block_id": "37afe017-5a04-4d87-96b0-b8f88a328d7a",
                    "description": "Product description",
                    "tag": "ProductType",
                    "status": "active",
                },
                "Example Product Block Two": {
                    "product_block_id": "e8312243-f5cc-4560-adf0-63be4cefeccd",
                    "description": "Product description",
                    "tag": "ProductType",
                    "status": "active",
                }
            }
    """
    uuids = {}
    for name, product_block in new.items():
        product_block["name"] = name
        uuids[name] = product_block["product_block_id"]
        conn.execute(
            sa.text(
                """
                INSERT INTO product_blocks (product_block_id, name, description, tag, status, created_at)
                VALUES (:product_block_id, :name, :description, :tag, :status, now())
                ON CONFLICT DO NOTHING;
                """
            ),
            product_block,
        )
        if "resource_types" in product_block:
            block_resource_types = {name: product_block["resource_types"]}
            create_resource_types_for_product_blocks(conn, block_resource_types)

        if "in_use_by_block_relations" in product_block:
            for in_use_by_block_name in product_block["in_use_by_block_relations"]:
                in_use_by_block_id = get_product_block_id_by_name(conn, in_use_by_block_name)
                add_product_block_relation_between_products_by_id(
                    conn, str(in_use_by_block_id), str(product_block["product_block_id"])
                )

        if "depends_on_block_relations" in product_block:
            for depends_on_block_name in product_block["depends_on_block_relations"]:
                depends_on_block_id = get_product_block_id_by_name(conn, depends_on_block_name)
                add_product_block_relation_between_products_by_id(
                    conn, str(product_block["product_block_id"]), str(depends_on_block_id)
                )

    return uuids


def create_resource_types_for_product_blocks(conn: sa.engine.Connection, new: dict) -> list[UUID]:
    """Create new resource types and link them to existing product_blocks by product_block name.

    Note: If the resource type already exists it will still add the resource type to the provided Product Blocks.

    Args:
        conn: DB connection as available in migration main file
        new: a dict with your product blocks and resource types
    Returns:
        List of resource type ids in order of insertion

    Usage examples:
        >>> new_stuff = {
            "ProductBlockName1": {
                "resource_type1": ("Resource description", "a47a3f96-c32f-4e4d-8e8c-11596451e878")
            },
            "ProductBlockName2": {
                "resource_type1": ("Resource description", "a47a3f96-c32f-4e4d-8e8c-11596451e878"),
                "resource_type2": ("Resource description", "dffe1890-e0f8-4ed5-8d0b-e769c3f726cc")
            }
        }
        >>> create_resource_types(conn, new_stuff)

    without extra ID's you don't need the tuple:

        >>> new_stuff = {
            "ProductBlockName1": {
                "resource_type1": "Resource description"
            },
            "ProductBlockName2": {
                "resource_type1": "Resource description",
                "resource_type1": "Resource description"
            }
        }

        >>> create_resource_types_for_product_blocks(conn, new_stuff)

    """
    insert_resource_type_with_id = sa.text(
        """INSERT INTO resource_types (resource_type_id, resource_type, description)
           VALUES (:resource_type_id, :resource_type, :description)
           ON CONFLICT DO NOTHING;"""
    )
    insert_resource_type_without_id = sa.text(
        """INSERT INTO resource_types (resource_type, description)
           VALUES (:resource_type, :description)
           ON CONFLICT DO NOTHING;"""
    )

    resource_type_ids = []
    for resource_types in new.values():
        for resource_type, values in resource_types.items():
            if isinstance(values, tuple):
                description, resource_type_id = values
                conn.execute(
                    insert_resource_type_with_id,
                    {
                        "resource_type_id": resource_type_id,
                        "resource_type": resource_type,
                        "description": description,
                    },
                )
            else:
                conn.execute(insert_resource_type_without_id, {"resource_type": resource_type, "description": values})
            resource_type_ids.append(get_resource_type_id_by_name(conn, resource_type))

    for product_block, resource_types in new.items():
        conn.execute(
            sa.text(
                """
                WITH resource_type_ids AS (SELECT resource_types.resource_type_id
                                           FROM resource_types
                                           WHERE resource_types.resource_type = ANY (:new_resource_types)),
                     service_attach_point AS (SELECT product_blocks.product_block_id
                                              FROM product_blocks
                                              WHERE product_blocks.name = :product_block_name)

                INSERT
                INTO product_block_resource_types (product_block_id, resource_type_id)
                SELECT service_attach_point.product_block_id,
                       resource_type_ids.resource_type_id
                FROM service_attach_point
                         CROSS JOIN
                     resource_type_ids
                """
            ),
            {
                "product_block_name": product_block,
                "new_resource_types": list(resource_types.keys()),
            },
        )
    return resource_type_ids


def create(conn: sa.engine.Connection, new: dict) -> None:  # noqa: C901
    """Call other functions in this file based on the schema.

    Args:
        conn: DB connection as available in migration main file
        new: a dict with everything you want to make and link

    Example:
        >>> new_stuff = {
                "products": {
                    "Example Product": {
                        "product_id": "c9dc2374-514c-11eb-b685-acde48001122",
                        "product_type": "ProductType1",
                        "description": "Product description",
                        "tag": "ProductType",
                        "status": "active",
                        "product_blocks": [
                            "Example Product Block"
                        ],
                        "fixed_inputs": {
                            "fixed_input_1": ("value", "f6a4f529-ad17-4ad8-b8ba-45684e2354ba"),
                            "fixed_input_2": ("value", "5a67321d-45d5-4921-aa93-b8708b5d74c6")
                        }
                    },
                    "Example Product 2": {
                        "product_id": "c9dc2374-514c-11eb-b685-acde48001122",
                        "product_type": "ProductType1",
                        "description": "Product description",
                        "tag": "ProductType",
                        "status": "active",
                        "product_block_ids": [
                            "37afe017-5a04-4d87-96b0-b8f88a328d7a"
                        ]
                    }
                },
                "product_blocks": {
                    "Example Product Block": {
                        "product_block_id": "37afe017-5a04-4d87-96b0-b8f88a328d7a",
                        "description": "Product description",
                        "tag": "ProductType",
                        "status": "active",
                        "resources": {
                            "resource_type1": ("Resource description", "a47a3f96-c32f-4e4d-8e8c-11596451e878"),
                            "resource_type2": ("Resource description", "dffe1890-e0f8-4ed5-8d0b-e769c3f726cc")
                        }
                    },
                    "Generated UUID Product Block": {
                        "product_block_id": "37afe017-5a04-4d87-96b0-b8f88a328d7a",
                        "description": "Product description",
                        "tag": "ProductType",
                        "status": "active",
                        "resources": {
                            "resource_type1": ("Resource description", "a47a3f96-c32f-4e4d-8e8c-11596451e878"),
                            "resource_type2": ("Resource description", "dffe1890-e0f8-4ed5-8d0b-e769c3f726cc")
                        }
                    }
                },
                "resources": {
                    "Existing Product": {
                        "resource_type1": ("Resource description", "a47a3f96-c32f-4e4d-8e8c-11596451e878"),
                        "resource_type2": ("Resource description", "dffe1890-e0f8-4ed5-8d0b-e769c3f726cc")
                    }
                },
                "workflows": {
                    "workflow_name": {
                        "workflow_id": "f2702074-3203-454c-b298-6dfa7675423d",
                        "target": "CREATE",
                        "description": "Workflow description",
                        "tag": "ProductType1",
                        "search_phrase": "Search Phrase%",
                    }
                }
            }
    """
    resources = new.get("resources", {})
    product_block_uuids = {}

    # Create defined product blocks
    if "product_blocks" in new:
        for product_block_name, product_block in new.get("product_blocks", {}).items():
            # Move resources into one dict
            if "resources" in product_block:
                res_dict = {product_block_name: product_block["resources"]}
                resources.update(res_dict)
                del product_block["resources"]
        product_block_uuids = create_product_blocks(conn, new["product_blocks"])

    def get_product_block_id(product_block_name: str) -> str:
        if product_block_id := product_block_uuids.get(product_block_name):
            return product_block_id

        try:
            return str(get_product_block_id_by_name(conn, product_block_name))
        except Exception:
            raise ValueError(f"{product_block_name} is not a valid product block.")

    # Create defined products
    if "products" in new:
        for _, product in new["products"].items():
            # The best practice is for a product to have only 1 root product block.
            # Migrations created through the generator adhere to this practice.
            product_block_ids = product.get("product_block_ids", [])
            if root_product_block := product.get("root_product_block"):
                root_product_block_id = get_product_block_id(root_product_block)
                if product_block_ids and product_block_ids != [root_product_block_id]:
                    logger.warning("Overriding hardcoded product_block_ids with root product block id")
                product["product_block_ids"] = [root_product_block_id]
                continue

            # To avoid forcing users to re-write old migrations or their products, this function will still
            # allow inserting multiple root product blocks.
            if "product_blocks" in product:
                product.setdefault("product_block_ids", [])
                for product_block_name in product["product_blocks"]:
                    product["product_block_ids"].append(get_product_block_id(product_block_name))
                del product["product_blocks"]
        create_products(conn, new["products"])

    # Create defined resource types
    if resources:
        create_resource_types_for_product_blocks(conn, resources)

    # Create defined workflows
    if "workflows" in new:
        create_workflows(conn, new["workflows"])

    # Ensure default workflows exist for all products
    ensure_default_workflows(conn)


def add_products_to_workflow_by_product_tag(
    conn: sa.engine.Connection, workflow_name: str, product_tag: str, product_name_like: str = "%%"
) -> None:
    """Add products to a workflow by product tag.

    Args:
        conn: DB connection as available in migration main file.
        workflow_name: Name of the workflow to add the products to.
        product_tag: Tag of the product to add to the workflow.
        product_name_like (optional): Part of the product name to get more specific products (necessary for fw v2)

    Usage:
    ```python
    product_tag = "product_tag"
    workflow_name = "workflow_name"
    add_products_to_workflow_by_product_tag(conn, product_tag, workflow_name)
    ```
    """

    conn.execute(
        sa.text(
            """
            WITH workflow AS (SELECT workflow_id
                              FROM workflows
                              WHERE name = :workflow_name)
            INSERT
            INTO products_workflows (product_id, workflow_id)
            SELECT p.product_id,
                   nw.workflow_id
            FROM products AS p
                     CROSS JOIN workflow AS nw
            WHERE p.tag = :product_tag
              AND name LIKE :product_name_like
            """
        ),
        {
            "workflow_name": workflow_name,
            "product_tag": product_tag,
            "product_name_like": product_name_like,
        },
    )


def remove_products_from_workflow_by_product_tag(
    conn: sa.engine.Connection, workflow_name: str, product_tag: str, product_name_like: str = "%%"
) -> None:
    """Delete products from a workflow by product tag.

    Args:
        conn: DB connection as available in migration main file.
        workflow_name: Name of the workflow that the products need to be removed from.
        product_tag: Tag of the product to remove from the workflow.
        product_name_like (optional): Part of the product name to get more specific products (necessary for fw v2)

    Usage:
    ```python
    product_tag = "product_tag"
    workflow_name = "workflow_name"
    remove_products_from_workflow_by_product_tag(conn, product_tag, workflow_name)
    ```
    """

    conn.execute(
        sa.text(
            """
            DELETE
            FROM products_workflows
            WHERE workflow_id = (SELECT workflow_id
                                 FROM workflows
                                 where name = :workflow_name)
              AND product_id IN (SELECT product_id
                                 FROM products
                                 WHERE tag = :product_tag
                                   AND name LIKE :product_name_like)
            """
        ),
        {
            "workflow_name": workflow_name,
            "product_tag": product_tag,
            "product_name_like": product_name_like,
        },
    )


def add_product_block_relation_between_products_by_id(
    conn: sa.engine.Connection, in_use_by_id: UUID | UUIDstr, depends_on_id: UUID | UUIDstr
) -> None:
    """Add product block relation by product block id.

    Args:
        conn: DB connection as available in migration main file.
        in_use_by_id: ID of the product block that uses another product block.
        depends_on_id: ID of the product block that is used as dependency.

    Usage:
    ```python
    in_use_by_id = "in_use_by_id"
    depends_on_id = "depends_on_id"
    add_product_block_relation_between_products_by_id(conn, in_use_by_id, depends_on_id)
    ```
    """

    conn.execute(
        sa.text(
            """
            INSERT INTO product_block_relations (in_use_by_id, depends_on_id)
            VALUES (:in_use_by_id, :depends_on_id)
            """
        ),
        {
            "in_use_by_id": in_use_by_id,
            "depends_on_id": depends_on_id,
        },
    )


def remove_product_block_relation_between_products_by_id(
    conn: sa.engine.Connection, in_use_by_id: UUID | UUIDstr, depends_on_id: UUID | UUIDstr
) -> None:
    """Remove product block relation by id.

    Args:
        conn: DB connection as available in migration main file.
        in_use_by_id: ID of the product block that uses another product block.
        depends_on_id: ID of the product block that is used as dependency.

    Usage:
        >>> in_use_by_id = "in_use_by_id"
        >>> depends_on_id = "depends_on_id"
        >>> remove_product_block_relation_between_products_by_id(
            conn, in_use_by_id, depends_on_id
        )
    """

    conn.execute(
        sa.text(
            """
            DELETE
            FROM product_block_relations
            WHERE in_use_by_id = :in_use_by_id
              AND depends_on_id = :depends_on_id
            """
        ),
        {
            "in_use_by_id": in_use_by_id,
            "depends_on_id": depends_on_id,
        },
    )


def delete_resource_type_by_id(conn: sa.engine.Connection, id: UUID | UUIDstr) -> None:
    """Delete resource type by resource type id.

    Args:
        conn: DB connection as available in migration main file.
        id: ID of the resource type to delete.

    Usage:
    ```python
    resource_type_id = "id"
    delete_resource_type_by_id(conn, resource_type_id)
    ```
    """
    conn.execute(sa.text("DELETE FROM resource_types WHERE resource_type_id=:id"), {"id": id})


def delete_resource_types_from_product_blocks(conn: sa.engine.Connection, delete: dict) -> None:
    """Delete resource type from product blocks.

    Note: the resource_type itself will not be deleted.

    Args:
        conn: DB connection as available in migration main file
        delete: dict of product_blocks and resource_types names that you want to unlink

    Example:
        >>> obsolete_stuff = {
            "ProductBlockName1": {
                "resource_type1": "Resource description"
            },
            "ProductBlockName2": {
                "resource_type1": "Resource description",
                "resource_type1": "Resource description"
            }
        }
        >>> delete_resource_types_from_product_blocks(conn, obsolete_stuff)
    """
    for product_block_name, resource_types in delete.items():
        conn.execute(
            sa.text(
                """DELETE
                   FROM product_block_resource_types
                       USING resource_types
                   WHERE
                       product_block_id = (SELECT product_block_id FROM product_blocks WHERE name = :product_block_name)
                     AND resource_types.resource_type_id = product_block_resource_types.resource_type_id
                     AND resource_types.resource_type = ANY (:obsolete_resource_types)"""
            ),
            {
                "product_block_name": product_block_name,
                "obsolete_resource_types": list(resource_types.keys()),
            },
        )


def delete_resource_types(conn: sa.engine.Connection, delete: Iterable) -> None:
    """Delete a resource type and it's occurrences in product blocks.

    Args:
        conn: DB connection as available in migration main file
        delete: list of resource_type names you want to delete

    Example:
        >>> obsolete_stuff = ["name_1", "name_2"]
        >>> delete_resource_types(conn, obsolete_stuff)
    """
    conn.execute(
        sa.text(
            """DELETE
               FROM product_block_resource_types
                   USING resource_types
               WHERE resource_types.resource_type_id = product_block_resource_types.resource_type_id
                 AND resource_types.resource_type = ANY (:obsolete)"""
        ),
        {"obsolete": tuple(delete)},
    )
    conn.execute(sa.text("DELETE FROM resource_types WHERE resource_type in :obsolete;"), {"obsolete": list(delete)})


def delete_products_by_tag(conn: sa.engine.Connection, name: str) -> None:
    conn.execute(
        sa.text(
            """DELETE
                            FROM products p
                            WHERE p.name = :name"""
        ),
        {"name": name},
    )


def delete_product(conn: sa.engine.Connection, name: str) -> None:
    """Delete a product and it's occurrences in workflows and product_blocks.

    Args:
        conn: DB connection as available in migration main file
        name: a product name you want to delete

    Example:
        >>> obsolete_stuff = "name_1"
        >>> delete_product(conn, obsolete_stuff)
    """
    conn.execute(
        sa.text(
            """
            WITH deleted_p AS (
                DELETE FROM products WHERE name = :name
                    RETURNING product_id),
                 deleted_p_pb AS (
                     DELETE FROM product_product_blocks WHERE product_id IN (SELECT product_id FROM deleted_p)),
                 deleted_pb_rt AS (
                     DELETE FROM products_workflows WHERE product_id IN (SELECT product_id FROM deleted_p))
            SELECT *
            from deleted_p;
            """
        ),
        {"name": name},
    )


def delete_product_block(conn: sa.engine.Connection, name: str) -> None:
    """Delete a product block and it's occurrences in resource types and products.

    Args:
        conn: DB connection as available in migration main file
        name: a product_block name you want to delete

    Example:
        >>> obsolete_stuff = "name_1"
        >>> delete_product_block(conn, obsolete_stuff)
    """
    conn.execute(
        sa.text(
            """
            WITH deleted_pb AS (
                DELETE FROM product_blocks WHERE name = :name
                    RETURNING product_block_id),
                 deleted_p_pb AS (
                     DELETE FROM product_product_blocks WHERE product_block_id IN (SELECT product_block_id FROM deleted_pb)),
                 deleted_pb_rt AS (
                     DELETE FROM product_block_resource_types WHERE product_block_id IN (SELECT product_block_id FROM deleted_pb))
            SELECT *
            from deleted_pb;
            """
        ),
        {"name": name},
    )


def delete_workflow(conn: sa.engine.Connection, name: str) -> None:
    """Delete a workflow and its occurrences in products.

    Note: the cascading delete rules in postgres will also ensure removal from `products_workflows`.

    Args:
        conn: DB connection as available in migration main file
        name: a workflow name you want to delete

    Example:
        >>> obsolete_stuff = "name_1"
        >>> delete_workflow(conn, obsolete_stuff)
    """

    conn.execute(
        sa.text(
            """
            DELETE
            FROM workflows
            WHERE name = :name;
            """
        ),
        {"name": name},
    )


def delete_resource_type(conn: sa.engine.Connection, resource_type: str) -> None:
    """Delete a resource type and it's occurrences in product blocks and products.

    Args:
        conn: DB connection as available in migration main file
        resource_type: a resource_type name you want to delete

    Example:
        >>> resource_type = "name_1"
        >>> delete_product_block(conn, resource_type)
    """
    conn.execute(
        sa.text(
            """
            WITH deleted_pb AS (
                DELETE FROM resource_types WHERE resource_type = :resource_type
                    RETURNING resource_type_id),
                 deleted_pb_rt AS (
                     DELETE FROM product_block_resource_types WHERE resource_type_id IN (SELECT resource_type_id FROM deleted_pb))
            SELECT *
            from deleted_pb;
            """
        ),
        {"resource_type": resource_type},
    )


def delete(conn: sa.engine.Connection, obsolete: dict) -> None:
    """Delete multiple products, product_blocks, resources, and workflows.

    Args:
        conn: DB connection as available in migration main file
        obsolete: a dict with everything you want to delete

    Example:
        >>> obsolete = [
                "products": [
                    "Example Product",
                    "Example Product 2"
                ],
                "product_blocks": [
                    "Example Product Block",
                    "Example Product Block 2"
                ],
                "resources": [
                        "resource_type4,
                        "resource_type5"
                ],
                "workflows": [
                    "workflow_name_a",
                    "workflow_name_b",
                ]
            ]
    """
    if "resource_types" in obsolete:
        for res_type in obsolete["resource_types"]:
            delete_resource_type(conn, res_type)

    if "product_blocks" in obsolete:
        for product_block_name in obsolete["product_blocks"]:
            delete_product_block(conn, product_block_name)

    if "products" in obsolete:
        for product in obsolete["products"]:
            delete_product(conn, product)

    if "workflows" in obsolete:
        for workflow in obsolete["workflows"]:
            delete_workflow(conn, workflow)


#
# Helpers for migrating to the new domain model relations as introduced in 0.3.3
#


def convert_resource_type_relations_to_instance_relations(
    conn: sa.engine.Connection, resource_type_id: UUID | UUIDstr, domain_model_attr: str, cleanup: bool = True
) -> None:
    """Move resouce type relations to instance type relations using resource type id.

    Note: It removes the resource type relations after moving! (comment out the second execute to try without the removal)

    Args:
        conn: DB connection as available in migration main file.
        resource_type_id: ID of the resource type that you want to move to instance relations.
        domain_model_attr: Name of the domain model attribute that connects the product blocks together.
        cleanup: remove old resource type relations after the migrate?

    Usage:
        >>> resource_type_id = "id"
        >>> domain_model_attr = "domain_model_attr"
        >>> convert_resource_type_relation_to_instance_relations(
            conn, resource_type_id, domain_model_attr
        )
    """
    conn.execute(
        sa.text(
            """
            INSERT INTO subscription_instance_relations (in_use_by_id, depends_on_id, order_id, domain_model_attr)
            WITH dependencies AS (SELECT siv.value                    AS subscription_id,
                                         siv.subscription_instance_id AS in_use_by_instance_id,
                                         si.product_block_id
                                  FROM subscription_instance_values AS siv
                                           LEFT JOIN subscription_instances AS si
                                                     on siv.subscription_instance_id = si.subscription_instance_id
                                  WHERE siv.resource_type_id = :resource_type_id)
            SELECT in_use_by_instance_id                                        AS in_use_by_id,
                   sii.subscription_instance_id                                 AS depends_on_id,
                   (row_number() OVER (PARTITION BY in_use_by_instance_id) - 1) AS order_id,
                   :domain_model_attr                                           AS domain_model_attr
            FROM subscription_instances AS sii
                     INNER JOIN dependencies AS dep ON sii.subscription_id = uuid(dep.subscription_id)
            ON CONFLICT DO NOTHING
            """
        ),
        {
            "resource_type_id": resource_type_id,
            "domain_model_attr": domain_model_attr,
        },
    )

    if cleanup:
        conn.execute(
            sa.text(
                """
                DELETE
                FROM subscription_instance_values
                WHERE resource_type_id = :resource_type_id
                """
            ),
            {"resource_type_id": resource_type_id},
        )


def convert_instance_relations_to_resource_type_relations_by_domain_model_attr(
    conn: sa.engine.Connection, domain_model_attr: str, resource_type_id: UUID | UUIDstr, cleanup: bool = True
) -> None:
    """Move instance type relations to resouce type relations by domain model attribute.

    Note: It removes the instance relations after moving! Use the `cleanup` argument if you don't want this

    Args:
        conn: DB connection as available in migration main file.
        domain_model_attr: Name of the domain model attribute that connects the product blocks together.
        resource_type_id: ID of the resource type that you want to move the instance relations to.
        cleanup: remove old instance relations after the migrate?

    Usage:
        >>> domain_model_attr = "domain_model_attr"
        >>> resource_type_id = "id"
        >>> convert_instance_relations_to_resource_type_relations_by_domain_model_attr(
            conn, domain_model_attr, resource_type_id
        )
    """
    conn.execute(
        sa.text(
            """
            INSERT INTO subscription_instance_values (subscription_instance_id, resource_type_id, value)
            WITH instance_relations AS (SELECT in_use_by_id, depends_on_id
                                        FROM subscription_instance_relations
                                        WHERE domain_model_attr = :domain_model_attr)
            SELECT ir.in_use_by_id    AS subscription_instance_id,
                   :resource_type_id  AS resource_type_id,
                   si.subscription_id AS value
            from subscription_instances as si
                     inner join instance_relations as ir on si.subscription_instance_id = ir.depends_on_id
            """
        ),
        {
            "domain_model_attr": domain_model_attr,
            "resource_type_id": resource_type_id,
        },
    )

    if cleanup:
        conn.execute(
            sa.text("DELETE FROM subscription_instance_relations WHERE domain_model_attr=:attr"),
            {"attr": domain_model_attr},
        )


def backfill_resource_type_with_default(
    conn: sa.engine.Connection, resource_type_id: UUID, product_block_name: str, value: Any
) -> None:
    conn.execute(
        sa.text(
            """
            INSERT INTO subscription_instance_values (subscription_instance_id, resource_type_id, value)
            SELECT si.subscription_instance_id AS subscription_instance_id,
                   :resource_type_id           AS resource_type_id,
                   :value                      AS value
            FROM subscription_instances AS si
                     JOIN product_blocks pb on si.product_block_id = pb.product_block_id
            WHERE pb.name = :product_block_name;
            """
        ),
        {
            "resource_type_id": resource_type_id,
            "product_block_name": product_block_name,
            "value": value,
        },
    )


def remove_resource_type_from_subscription_instance(conn: sa.engine.Connection, resource_type_name: str) -> None:
    resource_type_id = get_resource_type_id_by_name(conn, resource_type_name)

    conn.execute(
        sa.text(
            """DELETE
                   FROM subscription_instance_values
                   WHERE resource_type_id = :resource_type_id"""
        ),
        {"resource_type_id": resource_type_id},
    )



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import array
import base64
from dataclasses import dataclass

from pydantic import BaseModel

from orchestrator.search.core.exceptions import InvalidCursorError
from orchestrator.search.schemas.parameters import BaseSearchParameters
from orchestrator.search.schemas.results import SearchResult


@dataclass
class PaginationParams:
    """Parameters for pagination in search queries."""

    page_after_score: float | None = None
    page_after_id: str | None = None
    q_vec_override: list[float] | None = None


def floats_to_b64(v: list[float]) -> str:
    a = array.array("f", v)
    return base64.urlsafe_b64encode(a.tobytes()).decode("ascii")


def b64_to_floats(s: str) -> list[float]:
    raw = base64.urlsafe_b64decode(s.encode("ascii"))
    a = array.array("f")
    a.frombytes(raw)
    return list(a)


class PageCursor(BaseModel):
    score: float
    id: str
    q_vec_b64: str

    def encode(self) -> str:
        """Encode the cursor data into a URL-safe Base64 string."""
        json_str = self.model_dump_json()
        return base64.urlsafe_b64encode(json_str.encode("utf-8")).decode("utf-8")

    @classmethod
    def decode(cls, cursor: str) -> "PageCursor":
        """Decode a Base64 string back into a PageCursor instance."""
        try:
            decoded_str = base64.urlsafe_b64decode(cursor).decode("utf-8")
            return cls.model_validate_json(decoded_str)
        except Exception as e:
            raise InvalidCursorError("Invalid pagination cursor") from e


async def process_pagination_cursor(cursor: str | None, search_params: BaseSearchParameters) -> PaginationParams:
    """Process pagination cursor and return pagination parameters."""
    if cursor:
        c = PageCursor.decode(cursor)
        return PaginationParams(
            page_after_score=c.score,
            page_after_id=c.id,
            q_vec_override=b64_to_floats(c.q_vec_b64),
        )
    if search_params.vector_query:
        from orchestrator.search.core.embedding import QueryEmbedder

        q_vec_override = await QueryEmbedder.generate_for_text_async(search_params.vector_query)
        return PaginationParams(q_vec_override=q_vec_override)
    return PaginationParams()


def create_next_page_cursor(
    search_results: list[SearchResult], pagination_params: PaginationParams, limit: int
) -> str | None:
    """Create next page cursor if there are more results."""
    has_next_page = len(search_results) == limit and limit > 0
    if has_next_page:
        last_item = search_results[-1]
        cursor_data = PageCursor(
            score=float(last_item.score),
            id=last_item.entity_id,
            q_vec_b64=floats_to_b64(pagination_params.q_vec_override or []),
        )
        return cursor_data.encode()
    return None



# Copyright 2019-2020 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections.abc import Callable
from http import HTTPStatus
from typing import Any
from uuid import UUID

import structlog
from celery.result import AsyncResult
from kombu.exceptions import ConnectionError, OperationalError
from sqlalchemy import select

from orchestrator import app_settings
from orchestrator.api.error_handling import raise_status
from orchestrator.db import ProcessTable, db
from orchestrator.services.processes import (
    SYSTEM_USER,
    can_be_resumed,
    create_process,
    delete_process,
    set_process_status,
)
from orchestrator.services.workflows import get_workflow_by_name
from orchestrator.workflow import ProcessStat, ProcessStatus
from pydantic_forms.types import State

logger = structlog.get_logger(__name__)


def _block_when_testing(task_result: AsyncResult) -> None:
    # Enables "Sync celery tasks. This will let the app wait until celery completes"
    if app_settings.TESTING:
        process_id = task_result.get()
        if not process_id:
            raise RuntimeError("Celery worker has failed to resume process")


def _celery_start_process(pstat: ProcessStat, user: str = SYSTEM_USER, **kwargs: Any) -> UUID:
    """Client side call of Celery."""
    from orchestrator.services.tasks import NEW_TASK, NEW_WORKFLOW, get_celery_task

    if not (wf_table := get_workflow_by_name(pstat.workflow.name)):
        raise_status(HTTPStatus.NOT_FOUND, "Workflow in Database does not exist")

    task_name = NEW_TASK if wf_table.is_task else NEW_WORKFLOW
    trigger_task = get_celery_task(task_name)
    try:
        result = trigger_task.delay(pstat.process_id, user)
        _block_when_testing(result)
        return pstat.process_id
    except (ConnectionError, OperationalError) as e:
        # If connection to Redis fails and process can't be started, we need to remove the created process
        logger.warning("Connection error when submitting task to Celery. Delete newly created process from database.")
        delete_process(pstat.process_id)
        raise e


def _celery_resume_process(
    process: ProcessTable,
    *,
    user: str | None = None,
    **kwargs: Any,
) -> bool:
    """Client side call of Celery."""
    from orchestrator.services.tasks import RESUME_TASK, RESUME_WORKFLOW, get_celery_task

    last_process_status = process.last_status

    task_name = RESUME_TASK if process.workflow.is_task else RESUME_WORKFLOW
    trigger_task = get_celery_task(task_name)

    _celery_set_process_status_resumed(process.process_id)

    try:
        result = trigger_task.delay(process.process_id, user)
        _block_when_testing(result)

        return process.process_id
    except (ConnectionError, OperationalError) as e:
        logger.warning(
            "Connection error when submitting task to celery. Resetting process status back",
            current_status=process.last_status,
            last_status=last_process_status,
        )
        set_process_status(process.process_id, last_process_status)
        raise e


def _celery_set_process_status_resumed(process_id: UUID) -> None:
    """Set the process status to RESUMED to show its waiting to be picked up by a worker.

    uses with_for_update to lock the subscription in a transaction, preventing other changes.
    rolls back transation and raises an exception when it can't change to RESUMED to prevent it from being added to the queue.

    Args:
        process_id: Process ID to fetch process from DB
    """
    stmt = select(ProcessTable).where(ProcessTable.process_id == process_id).with_for_update()

    result = db.session.execute(stmt)
    locked_process = result.scalar_one_or_none()

    if not locked_process:
        raise ValueError(f"Process not found: {process_id}")

    if can_be_resumed(locked_process.last_status):
        locked_process.last_status = ProcessStatus.RESUMED
        db.session.commit()
    else:
        db.session.rollback()
        raise ValueError(f"Process has incorrect status to resume: {locked_process.last_status}")


def _celery_validate(validation_workflow: str, json: list[State] | None) -> None:
    pstat = create_process(validation_workflow, user_inputs=json)
    _celery_start_process(pstat)


CELERY_EXECUTION_CONTEXT: dict[str, Callable] = {
    "start": _celery_start_process,
    "resume": _celery_resume_process,
    "validate": _celery_validate,
}
